/**************************************************************************/
/*                                                                        */
/*  This file is part of Frama-Clang                                      */
/*                                                                        */
/*  Copyright (C) 2012-2016                                               */
/*    CEA (Commissariat à l'énergie atomique et aux énergies              */
/*         alternatives)                                                  */
/*                                                                        */
/*  you can redistribute it and/or modify it under the terms of the GNU   */
/*  Lesser General Public License as published by the Free Software       */
/*  Foundation, version 2.1.                                              */
/*                                                                        */
/*  It is distributed in the hope that it will be useful,                 */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*  GNU Lesser General Public License for more details.                   */
/*                                                                        */
/*  See the GNU Lesser General Public License version 2.1                 */
/*  for more details (enclosed in the file LICENSE).                      */
/*                                                                        */
/**************************************************************************/

//
// Description:
//   Implementation of the ACSL++ parser.
//

#include "clang/Basic/Version.h"
#include "llvm/ADT/SmallString.h"
#include "clang/AST/ASTContext.h"
#include "clang/AST/DeclTemplate.h"
#include "clang/Lex/Preprocessor.h"
#include "clang/Sema/Sema.h"
#include "clang/Sema/Lookup.h"
#include "clang/Basic/TargetInfo.h"
#include "clang/Lex/MacroArgs.h"

#include "ACSLParser.h"

namespace Acsl {

void Parser::addBuiltinBindings() {
  {
    std::string ident = "\\is_finite";
    logic_info def =
      logic_info_cons(
        qualified_name_cons(NULL,strdup(ident.c_str())),
        true,
        NULL,NULL,
        opt_none(),
        cons_container(
          logic_arg_decl_cons(logic_type_Lfloat(FDOUBLE),NULL),NULL),
        logic_body_LBnone());
    _arguments.addOverloadedLogicFunctions(ident,def);
    def =
      logic_info_cons(
        qualified_name_cons(NULL,strdup(ident.c_str())),
        true,
        NULL,NULL,
        opt_none(),
        cons_container(
          logic_arg_decl_cons(logic_type_Lfloat(FFLOAT),NULL),NULL),
        logic_body_LBnone());
    _arguments.addOverloadedLogicFunctions(ident,def);
  }
}

void
Parser::Arguments::addErrorMessagesFromLexer()
{ std::list<Lexer::Error>& errorList = _lexer.errorList();
  std::list<Lexer::Error>::const_iterator iterEnd = errorList.end();
  for (std::list<Lexer::Error>::const_iterator iter = errorList.begin();
      iter != iterEnd; ++iter) {
    ++_countErrors;
    if (_errorMessages) {
      std::string message("[lexer error] ");
      message.append(iter->message);
      if (_errorMessages->empty()
          || (_errorMessages->back().getMessage() != message))
        _errorMessages->push_back(ErrorMessage(message,
              iter->filePos, iter->linePos, iter->columnPos));
      if (_doesStopOnError || !_errorMessages || _countErrors >= 20)
        return;
    };
  };
  errorList.clear();
}

void
Parser::Arguments::eatToken(ReadResult& result) {
  if (_stateLexer == SLMacroTokens) {
    while (!_macroTokensStack.empty() && _macroTokensStack.back().empty()) {
       _macroTokensStack.pop_back();
       if (_currentMacroArgumentsStack.back())
         _currentMacroArgumentsStack.back()
           ->destroy(_clangSema->getPreprocessor());
       _currentMacroArgumentsStack.pop_back();
    }
    if (!_macroTokensStack.empty()) {
      std::pair<unsigned, DLexer::AbstractToken*>&
        token = _macroTokensStack.back().front();
      _lexer.setToken(DLexer::Token(token.first, token.second));
      _macroTokensStack.back().pop_front();
      result = RRHasToken;
    }
    else {
      _usedMacros.clear();
      _stringsForToken.clear();
      _stateLexer = SLStandard;
    };
  }
  else
    assert(_stateLexer == SLStandard);
  if (_stateLexer == SLStandard) {
    result = _lexer.readToken(*_buffer, _position, _tokenLocation);
    if (_lexer.hasErrors())
      addErrorMessagesFromLexer();
  };
  if (result == RRHasToken) {
    DLexer::AbstractToken::Type type = _lexer.queryToken().getType();
    if (type == DLexer::AbstractToken::TComment)
      result = RRContinueLexing;
    else if (type == DLexer::AbstractToken::TIdentifier) {
      if (isPreprocessingDirective(((const DLexer::IdentifierToken&)
          _lexer.getContentToken()).content())) {
        std::string identifier = ((const DLexer::IdentifierToken&)
          _lexer.getContentToken()).content();
        clang::MacroInfo* macro = getMacro(identifier);
        if (_usedMacros.find(identifier) != _usedMacros.end())
          result = RRHasToken;
        else if (handleMacroExpandedIdentifier(identifier, macro,
              result))
          result = RRContinueLexing;
        else if (result == RRNeedChars)
          result = RRContinueLexing;
      };
    };
  };
}

bool
Parser::Arguments::isPreprocessingDirective(const std::string& name) const
{ clang::IdentifierInfo& identifierInfo
    = _clangSema->getPreprocessor().getIdentifierTable().get(name);
  return identifierInfo.hasMacroDefinition();
}

clang::MacroInfo*
Parser::Arguments::getMacro(const std::string& name) const {
  clang::IdentifierInfo& identifierInfo
    = _clangSema->getPreprocessor().getIdentifierTable().get(name);
  return _clangSema->getPreprocessor().getMacroInfo(&identifierInfo);
}

bool
Parser::Arguments::isNextPPTokenLParen() {
  assert(_buffer);
  Lexer::ReadResult result = Lexer::RRNeedChars;
  do {
    eatToken(result);
    if (result == RRHasToken) {
      DLexer::AbstractToken token = _lexer.queryToken();
      DLexer::AbstractToken::Type type = token.getType();
      if (type == DLexer::AbstractToken::TComment)
        result = RRContinueLexing;
      else {
        if (type == DLexer::AbstractToken::TOperatorPunctuator) {
          if (((const DLexer::OperatorPunctuatorToken&) token).getType()
              == DLexer::OperatorPunctuatorToken::TOpenParen)
            return true;
        };
        return false;
      };
    };
  } while (result == RRContinueLexing);
  return false;
}

DLexer::Token
Parser::Arguments::convertToken(const clang::Token& source) const {
  switch (source.getKind()) {
    case clang::tok::eof: // End of file.
    case clang::tok::eod: // End of preprocessing directive (end of line inside
                          //   a directive).
    case clang::tok::code_completion: // Code completion marker
      assert(false); // unimplemented
    case clang::tok::comment:
      { DLexer::CommentToken* result = new DLexer::CommentToken;
        result->content() = std::string(source.getLiteralData(),
            source.getLength());
        return DLexer::Token(result);
      };
    // C99 6.4.2: Identifiers.
    case clang::tok::identifier:          // abcde123
      { DLexer::IdentifierToken* result = new DLexer::IdentifierToken;
        clang::IdentifierInfo* identifierInfo = source.getIdentifierInfo();
        result->content() = std::string(identifierInfo->getNameStart(),
            identifierInfo->getLength());
        return DLexer::Token(result);
      };
    case clang::tok::raw_identifier:      // Used only in raw lexing mode.
      { DLexer::IdentifierToken* result = new DLexer::IdentifierToken;
        result->content() =
          std::string(source.getRawIdentifier(), source.getLength());
        return DLexer::Token(result);
      };

    // C99 6.4.4.1: Integer Constants
    // C99 6.4.4.2: Floating Constants
    case clang::tok::numeric_constant:
      // Clang does not distinguish between integer and floating literals:
      // we have to lex the constant properly ourselves.
      { llvm::SmallString<128> spellingBuffer;
        spellingBuffer.resize(source.getLength() + 1);
        bool isInvalid = false;
        std::string tokSpelling = _clangSema->getPreprocessor()
          .getSpelling(source, spellingBuffer, &isInvalid);
        if (isInvalid)
          return DLexer::Token();

        Acsl::Lexer lexer;
        Acsl::Lexer::ReadResult readResult;
        do {
          size_t position = 0;
          readResult = lexer.readToken(tokSpelling, position, _tokenLocation,
              true /* doesForce*/);
        } while (readResult <= Acsl::Lexer::RRContinueLexing);
        return lexer.extractToken();
      };

    // C99 6.4.4: Character Constants
    case clang::tok::char_constant:       // 'a'
    case clang::tok::wide_char_constant:  // L'b'
    // C++0x Character Constants
    case clang::tok::utf16_char_constant: // u'a'
    case clang::tok::utf32_char_constant: // U'a'
      // see clang::Sema::ActOnCharacterConstant
      { llvm::SmallString<16> charBuffer;
        bool isInvalid = false;
        std::string thisTok = _clangSema->getPreprocessor().getSpelling(source,
            charBuffer, &isInvalid);
        if (isInvalid)
          return DLexer::Token();
        Acsl::Lexer lexer;
        Acsl::Lexer::ReadResult readResult;
        do {
          size_t position = 0;
          readResult = lexer.readToken(thisTok, position, _tokenLocation,
              true /* doesForce*/);
        } while (readResult <= Acsl::Lexer::RRContinueLexing);
        return lexer.extractToken();
      };

    // C99 6.4.5: String Literals.
    case clang::tok::string_literal:      // "foo"
    case clang::tok::wide_string_literal: // L"foo"
    case clang::tok::angle_string_literal:// <foo>
    // C++0x String Literals.
    case clang::tok::utf8_string_literal: // u8"foo"
    case clang::tok::utf16_string_literal:// u"foo"
    case clang::tok::utf32_string_literal:// U"foo"
      // see clang::Sema::ActOnStringLiteral
      // [TODO] note that multiple string tokens should be converted into
      // one string token!
      { DLexer::StringLiteralToken* result = new DLexer::StringLiteralToken(
          std::string(source.getLiteralData(), source.getLength()));
        return DLexer::Token(result);
      };

    // C99 6.4.6: Punctuators.
    case clang::tok::l_square:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TOpenBracket));
    case clang::tok::r_square:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TCloseBracket));
    case clang::tok::l_paren:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TOpenParen));
    case clang::tok::r_paren:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TCloseParen));
    case clang::tok::l_brace:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TOpenBrace));
    case clang::tok::r_brace:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TCloseBrace));
    case clang::tok::period:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TPunct));
    case clang::tok::ellipsis:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TEllipsis));
    case clang::tok::amp:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TAmpersand));
    case clang::tok::ampamp:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TLogicalAnd));
    case clang::tok::ampequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TBitAndAssign));
    case clang::tok::star:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TStar));
    case clang::tok::starequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TMultAssign));
    case clang::tok::plus:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TPlus));
    case clang::tok::plusplus:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TPlusPlus));
    case clang::tok::plusequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TPlusAssign));
    case clang::tok::minus:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TMinus));
    case clang::tok::arrow:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TArrow));
    case clang::tok::minusminus:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TMinusMinus));
    case clang::tok::minusequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TMinusAssign));
    case clang::tok::tilde:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TTilda));
    case clang::tok::exclaim:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TNot));
    case clang::tok::exclaimequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TDifferent));
    case clang::tok::slash:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TDivide));
    case clang::tok::slashequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TDivideAssign));
    case clang::tok::percent:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TProcent));
    case clang::tok::percentequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TModuloAssign));
    case clang::tok::less:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TLess));
    case clang::tok::lessless:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TLeftShift));
    case clang::tok::lessequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TLessOrEqual));
    case clang::tok::lesslessequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TLeftShiftAssign));
    case clang::tok::greater:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TGreater));
    case clang::tok::greatergreater:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TRightShift));
    case clang::tok::greaterequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TGreaterOrEqual));
    case clang::tok::greatergreaterequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TRightShiftAssign));
    case clang::tok::caret:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TBitXor));
    case clang::tok::caretequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TBitXorAssign));
    case clang::tok::pipe:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TBitOr));
    case clang::tok::pipepipe:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TLogicalOr));
    case clang::tok::pipeequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TBitOrAssign));
    case clang::tok::question:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TQuery));
    case clang::tok::colon:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TColon));
    case clang::tok::semi:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TSemiColon));
    case clang::tok::equal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TAssign));
    case clang::tok::equalequal:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TEqual));
    case clang::tok::comma:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TComma));
    case clang::tok::hash:
    case clang::tok::hashhash:
    case clang::tok::hashat:
      assert(false); // unimplemented

    // C++ Support
    case clang::tok::periodstar:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TPunctStar));
    case clang::tok::arrowstar:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TArrowStar));
    case clang::tok::coloncolon:
      return DLexer::Token(DLexer::OperatorPunctuatorToken()
          .setType(DLexer::OperatorPunctuatorToken::TColonColon));

    // Objective C support.
    case clang::tok::at:
      assert(false); // unimplemented

    // CUDA support.
    case clang::tok::lesslessless:
    case clang::tok::greatergreatergreater: 
      assert(false); // unimplemented

    // C99 6.4.1: Keywords.  These turn into kw_* tokens.
    case clang::tok::kw_auto:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TAuto));
    case clang::tok::kw_break:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TBreak));
    case clang::tok::kw_case:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TCase));
    case clang::tok::kw_char:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TChar));
    case clang::tok::kw_const:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TConst));
    case clang::tok::kw_continue:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TContinue));
    case clang::tok::kw_default:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TDefault));
    case clang::tok::kw_do:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TDo));
    case clang::tok::kw_double:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TDouble));
    case clang::tok::kw_else:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TElse));
    case clang::tok::kw_enum:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TEnum));
    case clang::tok::kw_extern:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //    .setType(DLexer::KeywordToken::TExtern));
    case clang::tok::kw_float:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TFloat));
    case clang::tok::kw_for:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TFor));
    case clang::tok::kw_goto:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TGoto));
    case clang::tok::kw_if:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TIf));
    case clang::tok::kw_inline:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TInline));
    case clang::tok::kw_int:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TInt));
    case clang::tok::kw_long:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TLong));
    case clang::tok::kw_register:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TRegister));
    case clang::tok::kw_restrict:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TRestrict));
    case clang::tok::kw_return:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TReturn));
    case clang::tok::kw_short:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TShort));
    case clang::tok::kw_signed:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TSigned));
    case clang::tok::kw_sizeof:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TSizeof));
    case clang::tok::kw_static:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TStatic));
    case clang::tok::kw_struct:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TStruct));
    case clang::tok::kw_switch:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TSwitch));
    case clang::tok::kw_typedef:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TTypedef));
    case clang::tok::kw_union:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TUnion));
    case clang::tok::kw_unsigned:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TUnsigned));
    case clang::tok::kw_void:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TVoid));
    case clang::tok::kw_volatile:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TVolatile));
    case clang::tok::kw_while:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TWhile));
    case clang::tok::kw__Alignas:
    case clang::tok::kw__Alignof:
    case clang::tok::kw__Atomic:
    case clang::tok::kw__Bool:
    case clang::tok::kw__Complex:
    case clang::tok::kw__Generic:
    case clang::tok::kw__Imaginary:
    case clang::tok::kw__Static_assert:
    case clang::tok::kw___func__:
    case clang::tok::kw___objc_yes:
    case clang::tok::kw___objc_no:
      assert(false); // unimplemented

    // C++ 2.11p1: Keywords.
    case clang::tok::kw_asm:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TAsm));
    case clang::tok::kw_bool:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TBool));
    case clang::tok::kw_catch:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TCatch));
    case clang::tok::kw_class:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TClass));
    case clang::tok::kw_const_cast:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TConstCast));
    case clang::tok::kw_delete:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TDelete));
    case clang::tok::kw_dynamic_cast:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TDynamicCast));
    case clang::tok::kw_explicit:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TExplicit));
    case clang::tok::kw_export:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TExport));
    case clang::tok::kw_false:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TFalse));
    case clang::tok::kw_friend:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TFriend));
    case clang::tok::kw_mutable:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TMutable));
    case clang::tok::kw_namespace:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TNamespace));
    case clang::tok::kw_new:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TNew));
    case clang::tok::kw_operator:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TOperator));
    case clang::tok::kw_private:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TPrivate));
    case clang::tok::kw_protected:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TProtected));
    case clang::tok::kw_public:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TPublic));
    case clang::tok::kw_reinterpret_cast:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TReinterpretCast));
    case clang::tok::kw_static_cast:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TStaticCast));
    case clang::tok::kw_template:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TTemplate));
    case clang::tok::kw_this:
      { DLexer::IdentifierToken* result = new DLexer::IdentifierToken;
         result->content() = "this";
         return DLexer::Token(result);
      }
    case clang::tok::kw_throw:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TThrow));
    case clang::tok::kw_true:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TTrue));
    case clang::tok::kw_try:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TTry));
    case clang::tok::kw_typename:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TTypename));
    case clang::tok::kw_typeid:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TTypeid));
    case clang::tok::kw_using:
      assert(false); // unimplemented
      // return DLexer::Token(DLexer::KeywordToken()
      //     .setType(DLexer::KeywordToken::TUsing));
    case clang::tok::kw_virtual:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TVirtual));
    case clang::tok::kw_wchar_t:
      return DLexer::Token(DLexer::KeywordToken()
          .setType(DLexer::KeywordToken::TWcharT));

    default:
      assert(false); // unimplemented
  }
  return DLexer::Token();
}

clang::Token
Parser::Arguments::convertToClang(DLexer::Token source) const {
  DLexer::AbstractToken token = source.getFullToken();
  switch (token.getType()) {
    case DLexer::AbstractToken::TIdentifier:
      { clang::Token identifierToken;
        identifierToken.startToken();
        identifierToken.clearFlag(clang::Token::NeedsCleaning);
        identifierToken.setIdentifierInfo(0);
        const std::string& identifier = ((const DLexer::IdentifierToken&)
              source.getContent()).content();
        // std::memset(&identifierToken, 0, sizeof(clang::Token));
        identifierToken.setLength(identifier.size());
        identifierToken.setLocation(_location /* .getLocWithOffset(...) */);
        identifierToken.setKind(clang::tok::raw_identifier);
        _stringsForToken.push_back(identifier);
        identifierToken.setRawIdentifierData(_stringsForToken.back().c_str());
        return identifierToken;
      };
    case DLexer::AbstractToken::TKeyword:
      { clang::Token result;
        // result.setLength(...);
        // result.setLocation(...);
        switch (((const DLexer::KeywordToken&) token).getType()) {
          case DLexer::KeywordToken::TAuto:
            result.setKind(clang::tok::kw_auto);
            break;
          case DLexer::KeywordToken::TBool:
            result.setKind(clang::tok::kw_bool);
            break;
          case DLexer::KeywordToken::TCase:
            result.setKind(clang::tok::kw_case);
            break;
          case DLexer::KeywordToken::TChar:
            result.setKind(clang::tok::kw_char);
            break;
          case DLexer::KeywordToken::TClass:
            result.setKind(clang::tok::kw_class);
            break;
          case DLexer::KeywordToken::TConst:
            result.setKind(clang::tok::kw_const);
            break;
          case DLexer::KeywordToken::TConstCast:
            result.setKind(clang::tok::kw_const_cast);
            break;
          case DLexer::KeywordToken::TDouble:
            result.setKind(clang::tok::kw_double);
            break;
          case DLexer::KeywordToken::TDynamicCast:
            result.setKind(clang::tok::kw_dynamic_cast);
            break;
          case DLexer::KeywordToken::TElse:
            result.setKind(clang::tok::kw_else);
            break;
          case DLexer::KeywordToken::TEnum:
            result.setKind(clang::tok::kw_enum);
            break;
          case DLexer::KeywordToken::TFalse:
            result.setKind(clang::tok::kw_false);
            break;
          case DLexer::KeywordToken::TFloat:
            result.setKind(clang::tok::kw_float);
            break;
          case DLexer::KeywordToken::TFor:
            result.setKind(clang::tok::kw_for);
            break;
          case DLexer::KeywordToken::TIf:
            result.setKind(clang::tok::kw_if);
            break;
          case DLexer::KeywordToken::TInt:
            result.setKind(clang::tok::kw_int);
            break;
          case DLexer::KeywordToken::TLong:
            result.setKind(clang::tok::kw_long);
            break;
          case DLexer::KeywordToken::TOperator:
            result.setKind(clang::tok::kw_operator);
            break;
          case DLexer::KeywordToken::TReinterpretCast:
            result.setKind(clang::tok::kw_reinterpret_cast);
            break;
          case DLexer::KeywordToken::TShort:
            result.setKind(clang::tok::kw_short);
            break;
          case DLexer::KeywordToken::TSigned:
            result.setKind(clang::tok::kw_signed);
            break;
          case DLexer::KeywordToken::TSizeof:
            result.setKind(clang::tok::kw_sizeof);
            break;
          case DLexer::KeywordToken::TStatic:
            result.setKind(clang::tok::kw_static);
            break;
          case DLexer::KeywordToken::TStaticCast:
            result.setKind(clang::tok::kw_static_cast);
            break;
          case DLexer::KeywordToken::TStruct:
            result.setKind(clang::tok::kw_struct);
            break;
          case DLexer::KeywordToken::TTrue:
            result.setKind(clang::tok::kw_true);
            break;
          case DLexer::KeywordToken::TTypeId:
            result.setKind(clang::tok::kw_typeid);
            break;
          case DLexer::KeywordToken::TTypename:
            result.setKind(clang::tok::kw_typename);
            break;
          case DLexer::KeywordToken::TUnion:
            result.setKind(clang::tok::kw_union);
            break;
          case DLexer::KeywordToken::TUnsigned:
            result.setKind(clang::tok::kw_unsigned);
            break;
          case DLexer::KeywordToken::TVirtual:
            result.setKind(clang::tok::kw_virtual);
            break;
          case DLexer::KeywordToken::TVoid:
            result.setKind(clang::tok::kw_void);
            break;
          case DLexer::KeywordToken::TVolatile:
            result.setKind(clang::tok::kw_volatile);
            break;
          case DLexer::KeywordToken::TWcharT:
            result.setKind(clang::tok::kw_wchar_t);
            break;
          case DLexer::KeywordToken::TWhile:
            result.setKind(clang::tok::kw_while);
            break;
          default:
            break;
        };
        return result;
      };
    case DLexer::AbstractToken::TLiteral:
      { // result.setLength(...);
        // result.setLocation(...);
        clang::Token result;
        result.startToken();
        result.clearFlag(clang::Token::NeedsCleaning);
        result.setIdentifierInfo(0);
        // std::memset(&result, 0, sizeof(clang::Token));
        // result.setLocation(getSourceLocation(BufferPtr, TokLen));

        switch (((const DLexer::LiteralToken&) token).getType()) {
          case DLexer::LiteralToken::TInteger:
            result.setKind(clang::tok::numeric_constant);    // 0x123
            { std::ostringstream out;
              out << ((const DLexer::IntegerLiteralToken&)
                  source.getContent()).getValue();
              _stringsForToken.push_back(out.str());
              result.setLength(_stringsForToken.back().size());
              result.setLiteralData(_stringsForToken.back().c_str());
            };
            break;
          case DLexer::LiteralToken::TCharacter:
            result.setKind(clang::tok::char_constant);       // 'a'
            { std::string value;
              value += ((const DLexer::CharacterLiteralToken&)
                  source.getContent()).getChar();
              _stringsForToken.push_back(value);
              result.setLength(_stringsForToken.back().size());
              result.setLiteralData(_stringsForToken.back().c_str());
            };
            break;
          case DLexer::LiteralToken::TFloating:
            result.setKind(clang::tok::numeric_constant);    // 0x123
            { std::ostringstream out;
              out << ((const DLexer::FloatingLiteralToken&)
                  source.getContent()).getValueAsString();
              _stringsForToken.push_back(out.str());
              result.setLength(_stringsForToken.back().size());
              result.setLiteralData(_stringsForToken.back().c_str());
            };
            break;
          case DLexer::LiteralToken::TString:
            result.setKind(clang::tok::string_literal);      // "foo"
            _stringsForToken.push_back(((const DLexer::StringLiteralToken&)
                  source.getContent()).content());
              result.setLength(_stringsForToken.back().size());
            result.setLiteralData(_stringsForToken.back().c_str());
            break;
          default:
            break;
        };
        return result;
      };
    case DLexer::AbstractToken::TOperatorPunctuator:
      { clang::Token result;
        switch (((const DLexer::OperatorPunctuatorToken&) token).getType()) {
          case DLexer::OperatorPunctuatorToken::TOpenBrace:
            result.setKind(clang::tok::l_brace);
            break;
          case DLexer::OperatorPunctuatorToken::TCloseBrace:
            result.setKind(clang::tok::r_brace);
            break;
          case DLexer::OperatorPunctuatorToken::TOpenBracket:
            result.setKind(clang::tok::l_square);
            break;
          case DLexer::OperatorPunctuatorToken::TCloseBracket:
            result.setKind(clang::tok::r_square);
            break;
          case DLexer::OperatorPunctuatorToken::TOpenParen:
            result.setKind(clang::tok::l_paren);
            break;
          case DLexer::OperatorPunctuatorToken::TCloseParen:
            result.setKind(clang::tok::r_paren);
            break;
          case DLexer::OperatorPunctuatorToken::TSemiColon:
            result.setKind(clang::tok::semi);
            break;
          case DLexer::OperatorPunctuatorToken::TColon:
            result.setKind(clang::tok::colon);
            break;
          case DLexer::OperatorPunctuatorToken::TEllipsis:
            result.setKind(clang::tok::ellipsis);
            break;
          case DLexer::OperatorPunctuatorToken::TQuery:
            result.setKind(clang::tok::question);
            break;
          case DLexer::OperatorPunctuatorToken::TColonColon:
            result.setKind(clang::tok::coloncolon);
            break;
          case DLexer::OperatorPunctuatorToken::TPunct:
            result.setKind(clang::tok::period);
            break;
          case DLexer::OperatorPunctuatorToken::TPunctStar:
            result.setKind(clang::tok::periodstar);
            break;
          case DLexer::OperatorPunctuatorToken::TComma:
            result.setKind(clang::tok::comma);
            break;
          case DLexer::OperatorPunctuatorToken::TArrowStar:
            result.setKind(clang::tok::arrowstar);
            break;
          case DLexer::OperatorPunctuatorToken::TArrow:
            result.setKind(clang::tok::arrow);
            break;
          case DLexer::OperatorPunctuatorToken::TPlus:
            result.setKind(clang::tok::plus);
            break;
          case DLexer::OperatorPunctuatorToken::TMinus:
            result.setKind(clang::tok::minus);
            break;
          case DLexer::OperatorPunctuatorToken::TStar:
            result.setKind(clang::tok::star);
            break;
          case DLexer::OperatorPunctuatorToken::TDivide:
            result.setKind(clang::tok::slash);
            break;
          case DLexer::OperatorPunctuatorToken::TProcent:
            result.setKind(clang::tok::percent);
            break;
          case DLexer::OperatorPunctuatorToken::TBitXor:
            result.setKind(clang::tok::caret);
            break;
          case DLexer::OperatorPunctuatorToken::TAmpersand:
            result.setKind(clang::tok::amp);
            break;
          case DLexer::OperatorPunctuatorToken::TBitOr:
            result.setKind(clang::tok::pipe);
            break;
          case DLexer::OperatorPunctuatorToken::TTilda:
            result.setKind(clang::tok::tilde);
            break;
          case DLexer::OperatorPunctuatorToken::TNot:
            result.setKind(clang::tok::exclaim);
            break;
          case DLexer::OperatorPunctuatorToken::TAssign:
            result.setKind(clang::tok::equal);
            break;
          case DLexer::OperatorPunctuatorToken::TLess:
            result.setKind(clang::tok::less);
            break;
          case DLexer::OperatorPunctuatorToken::TGreater:
            result.setKind(clang::tok::greater);
            break;
          case DLexer::OperatorPunctuatorToken::TPlusAssign:
            result.setKind(clang::tok::plusequal);
            break;
          case DLexer::OperatorPunctuatorToken::TMinusAssign:
            result.setKind(clang::tok::minusequal);
            break;
          case DLexer::OperatorPunctuatorToken::TMultAssign:
            result.setKind(clang::tok::starequal);
            break;
          case DLexer::OperatorPunctuatorToken::TDivideAssign:
            result.setKind(clang::tok::slashequal);
            break;
          case DLexer::OperatorPunctuatorToken::TModuloAssign:
            result.setKind(clang::tok::percentequal);
            break;
          case DLexer::OperatorPunctuatorToken::TBitXorAssign:
            result.setKind(clang::tok::caretequal);
            break;
          case DLexer::OperatorPunctuatorToken::TBitAndAssign:
            result.setKind(clang::tok::ampequal);
            break;
          case DLexer::OperatorPunctuatorToken::TBitOrAssign:
            result.setKind(clang::tok::pipeequal);
            break;
          case DLexer::OperatorPunctuatorToken::TLeftShift:
            result.setKind(clang::tok::lessless);
            break;
          case DLexer::OperatorPunctuatorToken::TRightShift:
            result.setKind(clang::tok::greatergreater);
            break;
          case DLexer::OperatorPunctuatorToken::TLeftShiftAssign:
            result.setKind(clang::tok::lesslessequal);
            break;
          case DLexer::OperatorPunctuatorToken::TRightShiftAssign:
            result.setKind(clang::tok::greatergreaterequal);
            break;
          case DLexer::OperatorPunctuatorToken::TEqual:
            result.setKind(clang::tok::equalequal);
            break;
          case DLexer::OperatorPunctuatorToken::TDifferent:
            result.setKind(clang::tok::exclaimequal);
            break;
          case DLexer::OperatorPunctuatorToken::TLessOrEqual:
            result.setKind(clang::tok::lessequal);
            break;
          case DLexer::OperatorPunctuatorToken::TGreaterOrEqual:
            result.setKind(clang::tok::greaterequal);
            break;
          case DLexer::OperatorPunctuatorToken::TLogicalAnd:
            result.setKind(clang::tok::ampamp);
            break;
          case DLexer::OperatorPunctuatorToken::TLogicalOr:
            result.setKind(clang::tok::pipepipe);
            break;
          case DLexer::OperatorPunctuatorToken::TPlusPlus:
            result.setKind(clang::tok::plusplus);
            break;
          case DLexer::OperatorPunctuatorToken::TMinusMinus:
            result.setKind(clang::tok::minusminus);
            break;
          default:
            break;
        };
        return result;
      };
    case DLexer::AbstractToken::TComment:
      { clang::Token commentToken;
        commentToken.startToken();
        commentToken.clearFlag(clang::Token::NeedsCleaning);
        commentToken.setIdentifierInfo(0);
        // std::memset(&commentToken, 0, sizeof(clang::Token));
        const std::string& comment = ((const DLexer::CommentToken&)
              source.getContent()).content();
        commentToken.setLength(comment.size());
        // commentToken.setLocation(getSourceLocation(BufferPtr, TokLen));
        commentToken.setKind(clang::tok::comment);
        _stringsForToken.push_back(comment);
        commentToken.setLiteralData(_stringsForToken.back().c_str());
        return commentToken;
      };
    default:
      break;
  };
  return clang::Token();
}

clang::MacroArgs*
Parser::Arguments::readFunctionLikeMacroArgs(const std::string& identifier,
    clang::MacroInfo *macro) {
  // clang comment: The number of fixed arguments to parse.
  unsigned numFixedArgsLeft = macro->getNumArgs();
  bool isVariadic = macro->isVariadic();

  // argTokens - Build up a list of tokens that make up each argument.  Each
  // argument is separated by an EOF token.  Use a SmallVector so we can avoid
  // heap allocations in the common case.
  llvm::SmallVector<clang::Token, 64> argTokens;
  Lexer::ReadResult parseResult = RRHasToken;

  unsigned numActuals = 0;
  DLexer::AbstractToken readToken = _lexer.queryToken();
  while (parseResult != Lexer::RRFinished
      && (readToken.getType() != DLexer::AbstractToken::TOperatorPunctuator
        || ((const DLexer::OperatorPunctuatorToken&) readToken).getType()
            != DLexer::OperatorPunctuatorToken::TCloseParen)) {
    assert(readToken.getType() == DLexer::AbstractToken::TOperatorPunctuator);
    DLexer::OperatorPunctuatorToken::Type firstTokenType =
      ((const DLexer::OperatorPunctuatorToken&) readToken).getType();
    assert(firstTokenType == DLexer::OperatorPunctuatorToken::TOpenParen
        || firstTokenType == DLexer::OperatorPunctuatorToken::TComma);
    // "l-paren-ness?", "only expect argument separators here"

    // C99 6.10.3p11: Keep track of the number of l_parens we have seen.  Note
    // that we already consumed the first one.
    unsigned numParens = 0;

    _lexer.clearToken();
    while (1) {
      // Read arguments as unexpanded tokens.  This avoids issues, e.g., where
      // an argument value in a macro could expand to ',' or '(' or ')'.
      do {
        eatToken(parseResult);
      } while (parseResult != Lexer::RRHasToken
          && parseResult != Lexer::RRFinished);
      readToken = _lexer.queryToken();

      if (parseResult == Lexer::RRFinished) // "#if f(<eof>" & "#if f(\n"
        break;
      else if (readToken.getType() == DLexer::AbstractToken::TOperatorPunctuator
        && ((const DLexer::OperatorPunctuatorToken&) readToken).getType()
            == DLexer::OperatorPunctuatorToken::TCloseParen) {
        // If we found the ) token, the macro arg list is done.
        if (numParens-- == 0)
          break;
      }
      else if (readToken.getType() == DLexer::AbstractToken::TOperatorPunctuator
        && ((const DLexer::OperatorPunctuatorToken&) readToken).getType()
            == DLexer::OperatorPunctuatorToken::TOpenParen)
        ++numParens;
      else if (numParens == 0
        && readToken.getType() == DLexer::AbstractToken::TOperatorPunctuator
        && ((const DLexer::OperatorPunctuatorToken&) readToken).getType()
            == DLexer::OperatorPunctuatorToken::TComma) {
        // Comma ends this argument if there are more fixed arguments expected.
        // However, if this is a variadic macro, and this is part of the
        // variadic part, then the comma is just an argument token.
        if (!isVariadic)
          break;
        if (numFixedArgsLeft > 1)
          break;
      }
      argTokens.push_back(convertToClang(_lexer.extractToken()));
    }

    // If this was an empty argument list foo(), don't add this as an empty
    // argument.
    if (argTokens.empty()
        && (readToken.getType() == DLexer::AbstractToken::TOperatorPunctuator
        && ((const DLexer::OperatorPunctuatorToken&) readToken).getType()
            == DLexer::OperatorPunctuatorToken::TCloseParen))
      break;

    // If this is not a variadic macro, and too many args were specified, emit
    // an error.
    if (!isVariadic && numFixedArgsLeft == 0)
      return 0;

    // Add a marker EOF token to the end of the token list for this argument.
    clang::Token EOFTok;
    EOFTok.startToken();
    EOFTok.setKind(clang::tok::eof);
    // EOFTok.setLocation(Tok.getLocation());
    EOFTok.setLength(0);
    argTokens.push_back(EOFTok);
    ++numActuals;
    assert(numFixedArgsLeft != 0 && "Too many arguments parsed");
    --numFixedArgsLeft;
  }

  // Okay, we either found the r_paren.  Check to see if we parsed too few
  // arguments.
  unsigned minArgsExpected = macro->getNumArgs();

  // See MacroArgs instance var for description of this.
  assert(numActuals == minArgsExpected);
  return clang::MacroArgs::create(macro, argTokens, false,
      _clangSema->getPreprocessor());
}

bool
Parser::Arguments::handleMacroExpandedIdentifier(const std::string& identifier,
    clang::MacroInfo *macro, ReadResult& parseResult) {
  bool isFirstExpansion = (_stateLexer == SLStandard);
  assert(isFirstExpansion || _stateLexer == SLMacroTokens);
  if (isFirstExpansion)
    assert(_usedMacros.empty() && _macroTokensStack.empty()
        && _currentMacroArgumentsStack.empty());

  clang::MacroArgs* macroArgs = NULL;
  // clang comments: If this is a function-like macro, read the arguments.
  if (macro->isFunctionLike()) {
    // clang comments: C99 6.10.3p10: If the preprocessing token immediately
    //   after the macro name isn't a '(', this macro should not be expanded.
    if (!isNextPPTokenLParen()) { // [TODO] without consuming the token
      assert(parseResult == RRHasToken);
      return true;
    };

    // clang comments: Remember that we are now parsing the arguments to a
    //   macro invocation. Preprocessor directives used inside macro arguments
    //   are not portable, and this enables the warning.
    _stateLexer = SLMacroArgs;
    macroArgs = readFunctionLikeMacroArgs(identifier, macro);
    // clang comments: Finished parsing args.
    _stateLexer = SLMacroTokens;

    // clang comments: If there was an error parsing the arguments, bail out.
    if (macroArgs == 0) {
      _stateLexer = isFirstExpansion ? SLStandard : SLMacroTokens;
      _lexer.clearToken();
      return false;
    };
  }
  // else
  //   loadNextToken();

  // clang comments: If we started lexing a macro, enter the macro expansion
  //   body. If this macro expands to no tokens, don't bother to push it onto
  //   the expansion stack, only to take it right back off.
  if (macro->getNumTokens() == 0) {
    // clang comments: No need for arg info.
    if (macroArgs)
      macroArgs->destroy(_clangSema->getPreprocessor());
    parseResult = RRNeedChars;
    _stateLexer = isFirstExpansion ? SLStandard : SLMacroTokens;
    return true;
  }
  else if (macro->getNumTokens() == 1 && !macroArgs &&
           (!macro->getReplacementToken(0).isAnyIdentifier() ||
            !macro->getReplacementToken(0)
              .getIdentifierInfo()->hasMacroDefinition())) {
    // clang comments: Otherwise, if this macro expands into a single
    //   trivially-expanded token: expand it now.
    // This handles common cases like "#define VAL 42".
    // We make an exception for macros that expand to other macros
    // in order to let the generic mechanism do the recursive expansion

    // clang comments: Replace the result token.
    _lexer.clearToken();
    _lexer.setToken(convertToken(macro->getReplacementToken(0)));
    _stateLexer = isFirstExpansion ? SLStandard : SLMacroTokens;
    return false;
  }

  // clang comments: Notice that this macro has been used.
  _usedMacros.insert(identifier);
  _macroTokensStack.push_back(std::list<std::pair<unsigned,
      DLexer::AbstractToken*> >());
  _currentMacroArgumentsStack.push_back(macroArgs);

  int numTokens = macro->getNumTokens();
  // clang comments: Start expanding the macro.
  clang::Token identifierToken;
  identifierToken.startToken();
  identifierToken.clearFlag(clang::Token::NeedsCleaning);
  identifierToken.setIdentifierInfo(0);
  // std::memset(&identifierToken, 0, sizeof(clang::Token));
  identifierToken.setLength(identifier.size());
  identifierToken.setLocation(_location /* .getLocWithOffset(...) */);
  identifierToken.setKind(clang::tok::raw_identifier);
  _stringsForToken.push_back(identifier);
  identifierToken.setRawIdentifierData(_stringsForToken.back().c_str());
  clang::Token macroValue;
  if (numTokens > 0) {
    assert(macro->isEnabled());

    // [TODO] verify the absence of memory leaks
    llvm::StringRef afterMacro;
    std::unique_ptr<llvm::MemoryBuffer> buff =
        llvm::MemoryBuffer::getMemBufferCopy(afterMacro, "<macros>");
    assert(buff && "Cannot create empty buffer for macro expansion");
    clang::SourceManager& source_manager
        = _clangSema->getPreprocessor().getSourceManager();
    clang::FileID file =
      source_manager.createFileID(
            std::move(buff),clang::SrcMgr::C_User,0,0,_location);
    assert(!file.isInvalid() && "Could not create FileID for macros?");
    _clangSema->getPreprocessor().EnterSourceFile(file, 0, _location);            

    _clangSema->getPreprocessor().EnterMacro(identifierToken, _location,
        macro, _currentMacroArgumentsStack.back());
    _clangSema->getPreprocessor().Lex(macroValue);
    while (macroValue.getKind() != clang::tok::eof) {
      DLexer::Token token = convertToken(macroValue);
      if (token.hasContent())
        _macroTokensStack.back().push_back(std::make_pair((unsigned)
              token.getType(), token.extractContent()));
      else
        _macroTokensStack.back().push_back(std::make_pair((unsigned)
              token.getType(), (DLexer::AbstractToken*) NULL));
      macroValue = clang::Token();
      _clangSema->getPreprocessor().Lex(macroValue);
    }
    _clangSema->getPreprocessor().EndSourceFile();
    _stateLexer = SLMacroTokens;
  };
  _lexer.clearToken();
  return true;
}

const clang::NamedDecl*
Parser::Arguments::isCodeName(const std::string& name,
    const clang::TemplateArgument** templateArgument, bool* hasOverLoad) const
{ if (name.length() == 0)
    return NULL;
#if 0
  if (_clangContext->isFunctionOrMethod()) {
    // Apparently, LookupName does not consider formals in this case...
    const clang::FunctionDecl* f =
      static_cast<const clang::FunctionDecl*>(_clangContext);
    clang::FunctionDecl::param_const_iterator iter = f->param_begin();
    const clang::NamedDecl* param = NULL;
    while (iter!=f->param_end()) {
      const std::string s = (*iter)->getNameAsString();
      if (s == name) {
        param = *iter;
        break;
      }
      iter++;
    }
    if (param) return param;
  }
#endif

#if 0
  // The method clang::DeclContext::lookup seems to fail to find
  // adequate declaration. 
  // std::pair<clang::IdentifierInfo,const unsigned char*> mem;
  // mem.second = (const unsigned char*) name.c_str();
  // clang::DeclarationName delcName(&mem.first);
  
  const clang::DeclContext* currentContext = _clangContext;
  while (currentContext != NULL) {
    // clang::DeclContext::lookup_const_result cdeclResults
    //   = currentContext->lookup(delcName);
    if (currentContext->isRecord()
        && static_cast<const clang::RecordDecl*>(currentContext)->getName()
            .str() == name)
      return static_cast<const clang::RecordDecl*>(currentContext);
    if (currentContext->isNamespace()
        && static_cast<const clang::NamespaceDecl*>(currentContext)->getName()
            .str() == name)
      return static_cast<const clang::NamespaceDecl*>(currentContext);
    clang::DeclContext::decl_iterator iterEnd = currentContext->decls_end();
    for (clang::DeclContext::decl_iterator iter = currentContext->decls_begin();
        iter != iterEnd; ++iter) {
      clang::Decl* decl = *iter;
      clang::Decl::Kind kind = decl->getKind();
      if (kind >= clang::Decl::firstNamed && kind <= clang::Decl::lastNamed) {
        assert(llvm::dyn_cast<clang::NamedDecl>(decl));
        if (static_cast<const clang::NamedDecl*>(decl)->getNameAsString()
            == name)
          return static_cast<const clang::NamedDecl*>(decl);
      };
    };
    // if (cdeclResults.first != cdeclResults.second) {
    //   if (hasOverLoad)
    //     *hasOverLoad = (cdeclResults.first+1 != cdeclResults.second);
    //   return *cdeclResults.first;
    // };
    currentContext = currentContext->getParent();
  };
  // see Sema::CppLookupName(LookupResult &R, Scope *S)
  // for a conform lookup function
#endif

#if 1
  // clang::Token identifierToken;
  // identifierToken.startToken();
  // identifierToken.clearFlag(clang::Token::NeedsCleaning);
  // identifierToken.setIdentifierInfo(0);
  // // std::memset(&identifierToken, 0, sizeof(clang::Token));
  // identifierToken.setLength(name.size());
  // // identifierToken.setLocation(getSourceLocation(BufferPtr, TokLen));
  // identifierToken.setKind(clang::tok::raw_identifier);
  // identifierToken.setRawIdentifierData(name.c_str());
  // clang::IdentifierInfo& identifierInfo = *_clangSema->getPreprocessor()
  //     .LookUpIdentifierInfo(identifierToken);
  clang::IdentifierInfo& identifierInfo
      = _clangSema->getPreprocessor().getIdentifierTable().get(name);
  clang::LookupResult R(*_clangSema, &identifierInfo
      /* &_clangAST->Idents.get(name) */,
      _location, clang::Sema::LookupOrdinaryName);
  // Argh! The method getScopeForContext uses getCurScope that contains
  // the following comment:
  //   This routine must only be used when it is certain that semantic analysis
  //   and the parser are in precisely the same context, which is not the case
  //   when, e.g., we are performing any kind of template instantiation.
  //   Therefore, the only safe places to use this scope are in the parser
  //   itself and in routines directly invoked from the parser and *never* from
  //   template substitution or instantiation.
  // clang::Scope* scope = _clangSema->getScopeForContext(
  //    const_cast<clang::DeclContext*>(_clangContext));
  if (_clangSema->LookupName(R, _clangScope /* scope */, false)) {
    if (R.getResultKind() == clang::LookupResult::Found)
      return R.getFoundDecl();
    if (hasOverLoad && R.getResultKind()
        == clang::LookupResult::FoundOverloaded)
      *hasOverLoad = true;
  }
  else {
    const clang::NamedDecl* result = NULL;
    const clang::TemplateArgument*
        res = _clang_utils->findTemplateArgument(name, result);
    if (res) {
      if (templateArgument)
        *templateArgument = res;
      return result;
    };
  };
#endif
  return NULL;
}

const clang::NamedDecl*
Parser::Arguments::findQualifiedName(const std::string& name,
    const clang::DeclContext* context, bool* hasOverload) const
{ if (name.length() == 0)
    return NULL;
  clang::LookupResult R(*_clangSema, &_clangAST->Idents.get(name), _location,
      clang::Sema::LookupOrdinaryName);
  if (_clangSema->LookupQualifiedName(R, const_cast<clang::DeclContext*>(
          context))) {
    if (R.getResultKind() == clang::LookupResult::Found)
      return R.getFoundDecl();
    if (hasOverload
        && R.getResultKind() == clang::LookupResult::FoundOverloaded)
      *hasOverload = true;
  };
  return NULL;
}

template <class TypeTraits>
typename TypeTraits::ResultType
Parser::Arguments::isTypedefType(TypeTraits& traits, qualified_name name) const
{ /* qualification */ list prequalification = name->prequalification;
  const clang::DeclContext*
    declContext = _clangSema->Context.getTranslationUnitDecl();
  while (prequalification) {
    std::string name;
    qualification qualif = (qualification) prequalification->element.container;
    if (qualif->tag_qualification == QNAMESPACE)
      name = qualif->cons_qualification.QNamespace.name;
    else if (qualif->tag_qualification == QSTRUCTORCLASS)
      name = qualif->cons_qualification.QStructOrClass.name;
    else if (qualif->tag_qualification == QTEMPLATEINSTANCE)
      name = qualif->cons_qualification.QTemplateInstance.name;
    else
      { assert(false); }

    clang::LookupResult R(*_clangSema, &_clangAST->Idents.get(name),
        _location, clang::Sema::LookupOrdinaryName);
    const clang::NamedDecl* found = NULL;
    if (!_clangSema->LookupQualifiedName(R,
          const_cast<clang::DeclContext*>(declContext))) {
      if (name.compare(0, 10, "anonymous_") == 0)
        found = _clang_utils->findAnonymousDecl(name);
      if (!found)
        return TypeTraits::False();
    };
    if (!found && !R.isSingleResult()) // FIXME
      return TypeTraits::False();
    if (!found)
      found = R.getFoundDecl();
    clang::Decl::Kind kind = found->getKind();
    if (kind == clang::Decl::Namespace) {
      assert(llvm::dyn_cast<clang::NamespaceDecl>(found));
      declContext = static_cast<const clang::NamespaceDecl*>(found);
    }
    else if (kind >= clang::Decl::firstTag && kind <= clang::Decl::lastTag) {
      assert(llvm::dyn_cast<clang::TagDecl>(found));
      const clang::TagDecl*
        tagContext = static_cast<const clang::TagDecl*>(found);
      declContext = tagContext;
      if (qualif->tag_qualification == QTEMPLATEINSTANCE) {
        int existingInstancesNumber
          = tagContext->getNumTemplateParameterLists();
        for (int instance = 0; instance < existingInstancesNumber; ++instance) {
          // clang::TemplateParameterList* currentParameters
          //   = tagContext->getTemplateParameterList(instance);
          // TODO: check if qualification matches with currentParameters
          // TODO: find the actual instance of tagContext
          assert(false); // unimplemented
        };
        // TODO: A better alternative is to instantiate via Sema the
        //   CXXRecordDecl that corresponds to tagDecl.
      };
    }
    else if (kind == clang::Decl::ClassTemplate) {
      const clang::TemplateArgumentList* defaultTemplateArgs
          = _clang_utils->findInstanceArguments(found);
      assert(defaultTemplateArgs);
      void* InsertPos = NULL;
      clang::ClassTemplateSpecializationDecl* specialization =
        ((clang::ClassTemplateDecl*) const_cast<clang::NamedDecl*>(found))
        ->findSpecialization(defaultTemplateArgs->asArray(), InsertPos);
      if (!specialization)
         return TypeTraits::False();
      declContext = specialization;
    }
    else
      return TypeTraits::False();
    prequalification = prequalification->next;
  };

  clang::LookupResult R(*_clangSema, &_clangAST->Idents.get(name->decl_name),
      _location, clang::Sema::LookupOrdinaryName);
  const clang::NamedDecl* found = NULL;
  if (!_clangSema->LookupQualifiedName(R,
        const_cast<clang::DeclContext*>(declContext))) {
    if (strncmp(name->decl_name, "anonymous_", 10) == 0)
      found = _clang_utils->findAnonymousDecl(name->decl_name);
    if (!found)
      return TypeTraits::False();
  }
  if (!found && !R.isSingleResult()) // FIXME
    return TypeTraits::False();
  if (!found)
    found = R.getFoundDecl();
  clang::Decl::Kind kind = found->getKind();
  if (!(kind >= clang::Decl::firstType && kind <= clang::Decl::lastType)
      && kind != clang::Decl::ClassTemplate)
    return TypeTraits::False();
  if (kind == clang::Decl::ClassTemplate) {
    if (!traits.hasTemplateArgs())
      return TypeTraits::False();

    assert(llvm::dyn_cast<clang::ClassTemplateDecl>(found));
    /* template_parameter */ list templateArgs = traits.getTemplateArgs();
    const clang::ClassTemplateDecl* templateDecl
      = static_cast<const clang::ClassTemplateDecl*>(found);
    const clang::ClassTemplateSpecializationDecl* foundSpecialization = NULL;
    if (templateDecl) {
      clang::ClassTemplateDecl::spec_iterator iterEnd
        = const_cast<clang::ClassTemplateDecl*>(templateDecl)->spec_end();
      for (clang::ClassTemplateDecl::spec_iterator iter
          = const_cast<clang::ClassTemplateDecl*>(templateDecl)->spec_begin();
          !foundSpecialization && iter != iterEnd; ++iter) {
        tkind matchTemplateParameters =
          tkind_TTemplateInstance(
            _clang_utils->getTemplateExtension(
              tokenSourceLocation(), iter->getTemplateArgs()));
        struct _tkind templateParameters;
        templateParameters.tag_tkind = TTEMPLATEINSTANCE;
        templateParameters.cons_tkind.TTemplateInstance.parameters
          = templateArgs;
        if (tkind_equal(&templateParameters, matchTemplateParameters))
          foundSpecialization = *iter;
        free_tkind(matchTemplateParameters);
      };
    };
    // void* insertionPos = NULL;
    // const clang::ClassTemplateSpecializationDecl* foundSpecialization =
    //     templateDecl->findSpecialization(traits.getTemplateArgs(),
    //     traits.getNumTemplateArgs(), insertionPos);
    // if (!foundSpecialization)
    //   return TypeTraits::False();
    // return traits(foundSpecialization->getTypeForDecl());
    return traits(foundSpecialization->getTypeForDecl());
  };
  assert(llvm::dyn_cast<clang::TypeDecl>(found));
  const clang::TypeDecl* typeFound = static_cast<const clang::TypeDecl*>(found);
  const clang::Type* type = typeFound->getTypeForDecl();
  if (kind >= clang::Decl::firstTypedefName
      && kind <= clang::Decl::lastTypedefName)
    type = static_cast<const clang::TypedefNameDecl*>(typeFound)
      ->getUnderlyingType().getTypePtr();
  return traits(type);
}

class Parser::Arguments::IsIntegralTypeTraits {
private:
  const Clang_utils* _utils;

public:
  IsIntegralTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  typedef bool ResultType;
  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  static bool False() { return false; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->isIntegralType(type); }
};

class Parser::Arguments::IsSignedTypeTraits {
private:
  const Clang_utils* _utils;

public:
  IsSignedTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef bool ResultType;
  static bool False() { return false; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->isSignedType(type); }
};

class Parser::Arguments::IsArithmeticTypeTraits {
private:
  const Clang_utils* _utils;

public:
  IsArithmeticTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef bool ResultType;
  static bool False() { return false; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->isArithmeticType(type); }
};

class Parser::Arguments::IsFloatingTypeTraits {
private:
  const Clang_utils* _utils;

public:
  IsFloatingTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef bool ResultType;
  static bool False() { return false; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->isFloatingType(type); }
};

class Parser::Arguments::IsPointerTypeTraits {
private:
  const Clang_utils* _utils;

public:
  IsPointerTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef bool ResultType;
  static bool False() { return false; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->isPointerType(type); }
};

class Parser::Arguments::IsReferenceTypeTraits {
private:
  const Clang_utils* _utils;

public:
  IsReferenceTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef bool ResultType;
  static bool False() { return false; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->isReferenceType(type); }
};

class Parser::Arguments::IsArrayTypeTraits {
private:
  const Clang_utils* _utils;

public:
  IsArrayTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef bool ResultType;
  static bool False() { return false; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->isArrayType(type); }
};

class Parser::Arguments::IsCompoundTypeTraits {
private:
  const Clang_utils* _utils;
  tkind* _templateKind;

public:
  IsCompoundTypeTraits(const Clang_utils* utils, tkind* templateKind)
    : _utils(utils), _templateKind(templateKind) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef qualified_name ResultType;
  static qualified_name False() { return NULL; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->makeCompoundType(type, _templateKind); }
};

class Parser::Arguments::GetRecordTypeTraits {
private:
  const Clang_utils* _utils;

public:
  GetRecordTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef const clang::RecordType* ResultType;
  static const clang::RecordType* False() { return NULL; }
  ResultType operator()(const clang::Type* type) const
    { ResultType result = NULL;
      if (type->getTypeClass() == clang::Type::Typedef) {
        assert(llvm::dyn_cast<clang::TypedefType>(type));
        type = static_cast<const clang::TypedefType*>(type)->desugar()
          .getTypePtr();
      };
      if (type->getTypeClass() == clang::Type::Record) {
        assert(llvm::dyn_cast<clang::RecordType>(type));
        result = static_cast<ResultType>(type);
      }
      return result;
    }
};

class Parser::Arguments::LogicArithmeticPromotionTraits {
private:
  const Clang_utils* _utils;

public:
  LogicArithmeticPromotionTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef logic_type ResultType;
  static logic_type False() { return NULL; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->logicArithmeticPromotion(clang::SourceLocation(),type); }
};

class Parser::Arguments::MakePointedTypeTraits {
private:
  const Clang_utils* _utils;

public:
  MakePointedTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef logic_type ResultType;
  static logic_type False() { return NULL; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->makePointedType(clang::SourceLocation(),type); }
};

class Parser::Arguments::MakeReferencedTypeTraits {
private:
  const Clang_utils* _utils;

public:
  MakeReferencedTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef logic_type ResultType;
  static logic_type False() { return NULL; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->makeReferencedType(clang::SourceLocation(),type); }
};

class Parser::Arguments::MakeElementArrayTypeTraits {
private:
  const Clang_utils* _utils;

public:
  MakeElementArrayTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef logic_type ResultType;
  static logic_type False() { return NULL; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->makeElementArrayType(clang::SourceLocation(),type); }
};

class Parser::Arguments::MakeLogicTypeTraits {
private:
  const Clang_utils* _utils;

public:
  MakeLogicTypeTraits(const Clang_utils* utils) : _utils(utils) {}

  bool hasTemplateArgs() const { return false; }
  /* template_parameter */ list getTemplateArgs() const { return NULL; }

  typedef logic_type ResultType;
  static logic_type False() { return NULL; }
  ResultType operator()(const clang::Type* type) const
    { return _utils->makeLogicType(clang::SourceLocation(),type); }
};

bool
Parser::Arguments::isIntegralTypedefType(qualified_name name) const
{ IsIntegralTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

bool
Parser::Arguments::isSignedTypedefType(qualified_name name) const
{ IsSignedTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

bool
Parser::Arguments::isArithmeticTypedefType(qualified_name name) const
{ IsArithmeticTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

bool
Parser::Arguments::isFloatingTypedefType(qualified_name name) const
{ IsFloatingTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

bool
Parser::Arguments::isPointerTypedefType(qualified_name name) const
{ IsPointerTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

bool
Parser::Arguments::isReferenceTypedefType(qualified_name name) const
{ IsReferenceTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

bool
Parser::Arguments::isArrayTypedefType(qualified_name name) const
{ IsArrayTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

bool Parser::Arguments::is_wchar_signed() const {
  return
    clang::TargetInfo::isTypeSigned(_clangAST->getTargetInfo().getWCharType());
}

bool Parser::Arguments::is_char_signed() const {
  return _clangAST->getLangOpts().CharIsSigned;
}

qualified_name
Parser::Arguments::makeCompoundTypedefType(qualified_name name,
    tkind* templateKind) const
{ IsCompoundTypeTraits traits(_clang_utils, templateKind);
  return isTypedefType(traits, name);
}

logic_type
Parser::Arguments::makeTypeOfPointed(qualified_name name) const
{ MakePointedTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

logic_type
Parser::Arguments::makeTypeOfReferenced(qualified_name name) const
{ MakeReferencedTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

logic_type
Parser::Arguments::makeTypeOfArrayElement(qualified_name name) const
{ MakeElementArrayTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

const clang::RecordType*
Parser::Arguments::getRecordType(qualified_name name) const
{ GetRecordTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

GlobalContext::NestedContext*
Parser::Arguments::findLogicName(qualified_name context,
    const std::string& name, tkind* templateParameterContext) const {
  GlobalContext::NestedContext* logicContext
      = _clang_utils->globalAcslContext().findAbsolute(context);
  if (logicContext) {
    if (templateParameterContext && *templateParameterContext
        && (*templateParameterContext)->tag_tkind == TTEMPLATEINSTANCE) {
      if (logicContext->isQualification() && logicContext->asQualification()
            .hasTemplateRecordType())
        logicContext = logicContext->asQualification().findInstance(
          (*templateParameterContext)->cons_tkind.TTemplateInstance.parameters);
      else
        logicContext = NULL;
    };
  };
  return logicContext
    ? _clang_utils->globalAcslContext().find(name, logicContext)
    : NULL;
}

logic_type
Parser::Arguments::logicArithmeticPromotion(qualified_name name) const
{ LogicArithmeticPromotionTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

logic_type
Parser::Arguments::makeLogicType(qualified_name name) const
{ MakeLogicTypeTraits traits(_clang_utils);
  return isTypedefType(traits, name);
}

qualified_name
Parser::Arguments::makeLogicQualifiedName(const std::string& name) const {
  const clang::DeclContext* ctxt = _clangContext;
  return _clang_utils->makeQualifiedName(ctxt, copy_string(name));
}

logic_type
Parser::Arguments::queryThisType() const {
  const clang::DeclContext* currentContext = _clangContext;
  while ((currentContext != NULL) && !currentContext->isRecord())
    currentContext = currentContext->getParent();
  if (!currentContext)
    return NULL;
  /* qualification */ list prequalification = NULL;
  const clang::DeclContext* qualContext = currentContext->getParent();
  while (qualContext != NULL) {
    if (qualContext->isRecord()) {
      const clang::RecordDecl* decl = static_cast<const clang::RecordDecl*>
          (qualContext);
      qualification res = NULL;
      if (decl->getKind() == clang::Decl::ClassTemplateSpecialization) {
        const clang::ClassTemplateSpecializationDecl* TSD
            = llvm::dyn_cast<clang::ClassTemplateSpecializationDecl>(decl);
        res = qualification_QTemplateInstance(
          strdup(decl->getName().str().c_str()),
          _clang_utils->getTemplateExtension(
            tokenSourceLocation(),TSD->getTemplateArgs()));
      }
      else
        res = qualification_QStructOrClass(
          strdup(decl->getName().str().c_str()));
      prequalification = cons_container(res, prequalification);
    }
    else if (qualContext->isNamespace())
      prequalification = cons_container(qualification_QNamespace(
        strdup(static_cast<const clang::NamespaceDecl*>(qualContext)
          ->getName().str().c_str())), prequalification);
    qualContext = qualContext->getParent();
  };
  const clang::RecordDecl* decl = static_cast<const clang::RecordDecl*>
      (currentContext);
  tkind templateParameters = NULL;
  if (decl->getKind() == clang::Decl::ClassTemplateSpecialization) {
    const clang::ClassTemplateSpecializationDecl* TSD
        = llvm::dyn_cast<clang::ClassTemplateSpecializationDecl>(decl);
    templateParameters =
      tkind_TTemplateInstance(
        _clang_utils->getTemplateExtension(
          tokenSourceLocation(),TSD->getTemplateArgs()));
  }
  else
    templateParameters = tkind_TStandard();
  const clang::RecordDecl* record =
    static_cast<const clang::RecordDecl*>(currentContext);
  qualified_name rec_name =
    qualified_name_cons(
      prequalification, strdup(record->getName().str().c_str()));
  logic_type classType =
    record->isUnion() ?
    logic_type_Lunion(rec_name, templateParameters):
    logic_type_Lstruct(rec_name, templateParameters);
  if (isGlobalAnnotContext()) return classType;
  else return logic_type_Lpointer(classType);
}

term_lval Parser::Arguments::thisLval () const {
  logic_type thisType = queryThisType();
  if (!thisType) return NULL;
  term_lval res = NULL;
  if (thisType->tag_logic_type == LPOINTER) {
    term_lval thisPtr =
      term_lval_cons(
        term_lhost_TVar(
          logic_var_cons(qualified_name_cons(NULL, strdup("this")), LVFORMAL)),
        term_offset_TNoOffset());
    res =
      term_lval_cons(
        term_lhost_TMem(
          term_cons(term_node_TLval(thisPtr), newTokenLocation(), NULL)),
        term_offset_TNoOffset());
  } else {
    res = 
      term_lval_cons(
        term_lhost_TVar(
          logic_var_cons(qualified_name_cons(NULL,strdup("\\this")), LVFORMAL)),
        term_offset_TNoOffset());
  }
  free_logic_type(thisType);
  return res;
}

class Parser::Arguments::RetrieveTypeOfFieldTraits {
private:
  const Clang_utils* _utils;
  const RTTITable* _rttiTable;
  std::string _fieldName;
  term_offset& _offset;
  logic_type& _ltype;
  tkind _templateParameters;
  std::string _errorMessage;
  const clang::ASTContext* _clangAST;
  clang::Sema* _clangSema;
  clang::SourceLocation _location;

public:
  RetrieveTypeOfFieldTraits(const Clang_utils* utils,const RTTITable* rttiTable,
      const std::string& fieldName, term_offset& offset, logic_type& ltype,
      tkind templateParameters, const clang::ASTContext* clangAST,
      clang::Sema* clangSema, clang::SourceLocation location)
    : _utils(utils), _rttiTable(rttiTable), _fieldName(fieldName),
      _offset(offset), _ltype(ltype), _templateParameters(templateParameters),
      _clangAST(clangAST), _clangSema(clangSema), _location(location) {}

  bool hasTemplateArgs() const
    { return _templateParameters->tag_tkind == TTEMPLATEINSTANCE; }
  /* template_parameter */ list getTemplateArgs() const
    { return _templateParameters->cons_tkind.TTemplateInstance.parameters; }
  const std::string& errorMessage() const { return _errorMessage; }

  typedef bool ResultType;
  static bool False() { return false; }
  ResultType operator()(const clang::Type* type)
    { return _utils->retrieveTypeOfField(type, _fieldName, _offset, _ltype,
        _errorMessage, _clangAST, _clangSema, _location, *_rttiTable);
    }
};

bool
Parser::Arguments::retrieveTypeOfField(qualified_name name,
    tkind templateParameters, const std::string& fieldName,
    term_offset& offset, logic_type& ltype, std::string& errorMessage)
{ RetrieveTypeOfFieldTraits traits(_clang_utils, &_rttiTable, fieldName,
      offset, ltype, templateParameters, _clangAST, _clangSema, _location);
  bool result = isTypedefType(traits, name);
  if (!result)
    errorMessage = traits.errorMessage();
  return result;
}

bool
Parser::Arguments::addOverloadedLogicFunctionOrOperator(const std::string& name,
    logic_info info, bool isMethod,
    DLexer::OperatorPunctuatorToken::Type* codeOperator) {
  GlobalContext::NestedContext::SonsSet* localScope = NULL;
  if (_currentLogicScope) {
    localScope = _currentLogicScope->ssons();
    if (!localScope && _currentLogicScope->sparent())
      localScope = _currentLogicScope->sparent()->ssons();
  }
  if (!localScope)
    localScope = &_clang_utils->globalAcslContext().logicTable();
  GlobalContext::NestedContext locateContext(name);
  GlobalContext::NestedContext::SonsSet::iterator
    found = localScope->find(&locateContext);
  if (found == localScope->end()) {
    if (!codeOperator)
      localScope->insert(
          new GlobalContext::OverloadedLogicFunctions(name, info, isMethod));
    else
      localScope->insert(new GlobalContext::OverloadedLogicOperators(name,
            *codeOperator, info, isMethod));
  }
  else if (!(*found)->isOverloadedLogicFunctions())
    return false;
  else {
    typedef GlobalContext::OverloadedLogicFunctions::Functions Functions;
    const Functions& existingInfos
      = (*found)->asOverloadedLogicFunctions().getFunctions();
    Functions::const_iterator iterEnd = existingInfos.end();
    for (Functions::const_iterator iter = existingInfos.begin();
        iter != iterEnd; ++iter) {
      logic_info sourceInfo = iter->second;

      std::map<std::string, std::string> renaming;
      bool isEqual = true;
      if (sourceInfo->tparams) {
        if (!info->tparams)
          isEqual = false;
        /* const char* */ list infoParamNames = info->tparams,
                        sourceInfoParamNames = sourceInfo->tparams;
        while (isEqual && infoParamNames && sourceInfoParamNames) {
          renaming.insert(std::pair<std::string, std::string>(
            (const char*) sourceInfoParamNames->element.container,
            (const char*) infoParamNames->element.container));
          infoParamNames = infoParamNames->next;
          sourceInfoParamNames = sourceInfoParamNames->next;
        };
        if (infoParamNames || sourceInfoParamNames)
          isEqual = false;
      };
      isEqual = isEqual
        || (info->returned_type->is_some == sourceInfo->returned_type->is_some);
      if (isEqual && info->returned_type->is_some)
        isEqual = logic_type_equal(
            (logic_type) sourceInfo->returned_type->content.container,
            (logic_type) info->returned_type->content.container);
      if (isEqual) {
        /* logic_arg_decl */ list args = info->profile,
                                  sourceArgs = sourceInfo->profile;
        while (isEqual && args && sourceArgs) {
          isEqual = logic_type_equal(
            ((logic_arg_decl) args->element.container)->la_type,
            ((logic_arg_decl) sourceArgs->element.container)->la_type);
          args = args->next;
          sourceArgs = sourceArgs->next;
        };
        isEqual = isEqual && (args == NULL) && (sourceArgs == NULL);
      };
      if (isEqual) {
        if (sourceInfo->fun_body->tag_logic_body == LBNONE) {
          free_logic_body(sourceInfo->fun_body);
          sourceInfo->fun_body = info->fun_body;
          info->fun_body = logic_body_LBnone();
        }
        else if (info->fun_body->tag_logic_body == LBNONE)
          return false;
        if (!sourceInfo->arg_labels) {
          sourceInfo->arg_labels = info->arg_labels;
          info->arg_labels = NULL;
        }
        else {
          while (info->arg_labels) {
            /* logic_label */ list sourceLabelIter = sourceInfo->arg_labels;
            do {
              if (logic_label_equal(
                    (logic_label) sourceLabelIter->element.container,
                    (logic_label) info->arg_labels->element.container))
                break;
              sourceLabelIter = sourceLabelIter->next;
            } while (sourceLabelIter);
            if (!sourceLabelIter) {
              sourceInfo->arg_labels = cons_container((logic_label) info
                  ->arg_labels->element.container, sourceInfo->arg_labels);
              /* logic_label */ list cell = info->arg_labels;
              info->arg_labels = info->arg_labels->next;
              free(cell);
            }
            else {
              /* logic_label */ list cell = info->arg_labels;
              free_logic_label((logic_label)
                  info->arg_labels->element.container);
              info->arg_labels = info->arg_labels->next;
              free(cell);
            };
            info->arg_labels = info->arg_labels->next;
          };
        };
        return true;
      }
    };
    (*found)->asOverloadedLogicFunctions().addFunction(info, isMethod);
  };
  return true;
}
 
bool
Parser::Arguments::addLogicType(const std::string& name, logic_type_info
    ltypeInfo) {
  GlobalContext::NestedContext::SonsSet* localScope = NULL;
  if (_currentLogicScope) {
    localScope = _currentLogicScope->ssons();
    if (!localScope && _currentLogicScope->sparent())
      localScope = _currentLogicScope->sparent()->ssons();
  }
  if (!localScope)
    localScope = &_clang_utils->globalAcslContext().logicTable();
  GlobalContext::NestedContext locateContext(name);
  GlobalContext::NestedContext::SonsSet::iterator
    found = localScope->find(&locateContext);
  if (found != localScope->end())
    return false;

  localScope->insert(new GlobalContext::LogicType(name, ltypeInfo));
  return true;
}

GlobalContext::LogicType*
Parser::Arguments::findGlobalLogicType(qualified_name name) const {
  GlobalContext::NestedContext* result
    = _clang_utils->globalAcslContext().findAbsolute(name);
  return (result && result->isLogicType()) ? &result->asLogicType() : NULL;
}

/* qualification */ list
Parser::Arguments::createPrequalification() const {
  /* qualification */ list result = NULL;
  GlobalContext::NestedContext* scope = _currentLogicScope;
  while (scope) {
    if (scope->isQualification()) {
      result = cons_container(scope->asQualification().getQualification(),
          result);
      scope = scope->sparent();
    }
    else if (scope->isTemplateQualification()) {
      GlobalContext::NestedContext* rightParent = scope->sparent();
      assert(rightParent && rightParent->isQualification());
      result = cons_container(scope->asTemplateQualification()
          .getQualification(strdup(rightParent->asQualification()
            .getName().c_str())), result);
      scope = rightParent->sparent();
    }
    else
      assert(false);
  };
  return result;
}

GlobalContext::LogicVariable*
Parser::Arguments::findGlobalLogicVariable(qualified_name name) const {
  GlobalContext::NestedContext* result
    = _clang_utils->globalAcslContext().findAbsolute(name);
  return (result && result->isLogicVariable())
    ? &result->asLogicVariable() : NULL;
}

/* use Sema instead
 * class FoundResults {
   private:
    clang::DeclContext::lookup_const_result cdeclResults;
    Acsl::GlobalContext::NestedContext* pncLogicResults;

   public:
    FoundResults() : pncLogicResults(NULL) {}
    FoundResults(const FoundResults& source)
      : cdeclResults(source.cdeclResults),
        pncLogicResults(source.pncLogicResults) {}
   };
   
   FoundResults findIdentifier(const std::string& identifier,
      Acsl::GlobalContext::NestedContext* logicContext,
      clang::DeclContext* clangContext)
    { FoundResults result;
      Acsl::GlobalContext::NestedContext* logicResult = gcAcslContext.find(
        identifier, logicContext);
      if (!logicResult || !logicResult->isLogicVariable()) {
        std::pair<clang::IdentifierInfo,const unsigned char*> mem;
        mem.second = identifier.c_str();
        clang::DeclarationName name(&mem);
        result.cdeclResults = clangContext->lookup(name);
        if (result.cdeclResults)
      };
   }
*/

logic_label
Parser::Arguments::findLogicLabel(const std::string& name) const {
  std::set<std::string>::iterator found = _logicLabels.find(name);
  if (found != _logicLabels.end())
    return logic_label_LogicLabel(strdup(found->c_str()));

  clang::IdentifierInfo& identifierInfo
      = _clangSema->getPreprocessor().getIdentifierTable().get(name);
  clang::NamedDecl* foundDecl = _clangSema->LookupSingleName(
      _clangScope /* scope */, &identifierInfo, _location,
      clang::Sema::LookupLabel, clang::Sema::NotForRedeclaration);
  if (!foundDecl || (foundDecl->getKind() != clang::Decl::Label))
    return NULL;
  assert(llvm::dyn_cast<clang::LabelDecl>(foundDecl));
  return logic_label_StmtLabel(strdup(foundDecl->getName().str().c_str()));
}

logic_type
Parser::Arguments::createResultType() const {
  const clang::DeclContext* context = _clangContext;
  clang::Decl::Kind declKind;
  if (!_result_context) return NULL;
  while (context
      && !((declKind = context->getDeclKind()) >= clang::Decl::firstFunction
      && declKind <= clang::Decl::lastFunction))
    context = context->getParent();
  if (!context)
    return NULL;
  assert(llvm::dyn_cast<clang::FunctionDecl>(context));
  const clang::FunctionDecl*
    functionDecl = static_cast<const clang::FunctionDecl*>(context);
  const clang::SourceLocation loc = tokenSourceLocation();
  return
    _clang_utils->makeLogicType(loc,functionDecl->getReturnType().getTypePtr());
}

void
Parser::parse(const std::string& buffer) {
  _arguments.setBuffer(buffer, _errorMessages);
  _arguments._tokenLocation->linenum2 = _arguments._tokenLocation->linenum1;
  _arguments._tokenLocation->charnum2 = _arguments._tokenLocation->charnum1-1;
  ReadResult parseResult = _arguments.lexer().readToken(buffer,
      _arguments._position, _arguments._tokenLocation);
  while (parseResult != RRFinished) {
    if (parseResult == RRNeedChars) {
      if (_arguments._stateLexer == Arguments::SLMacroTokens) {
        while (!_arguments._macroTokensStack.empty()
             && _arguments._macroTokensStack.back().empty()) {
           _arguments._macroTokensStack.pop_back();
           if (_arguments._currentMacroArgumentsStack.back())
             _arguments._currentMacroArgumentsStack.back()
               ->destroy(_arguments._clangSema->getPreprocessor());
           _arguments._currentMacroArgumentsStack.pop_back();
        }
        if (!_arguments._macroTokensStack.empty()) {
          std::pair<unsigned, DLexer::AbstractToken*>&
            token = _arguments._macroTokensStack.back().front();
          _arguments.lexer().setToken(DLexer::Token(token.first, token.second));
          _arguments._macroTokensStack.back().pop_front();
          parseResult = RRHasToken;
        }
        else {
          _arguments._usedMacros.clear();
          _arguments._stringsForToken.clear();
          _arguments._stateLexer = Arguments::SLStandard;
        };
      }
      else
        assert(_arguments._stateLexer == Arguments::SLStandard);
      if (_arguments._stateLexer == Arguments::SLStandard) {
        parseResult = _arguments.lexer().readToken(buffer, _arguments._position,
            _arguments._tokenLocation, true);
        if (_arguments.lexer().hasErrors())
          _arguments.addErrorMessagesFromLexer();
      };
      if (parseResult == RRFinished)
        break;
      if (parseResult == RRNeedChars)
        break; // could add characters to buffer
    };
 
    if (!_errorMessages.empty()) {
      std::list<ErrorMessage>::const_iterator iter = _errorMessages.begin();
      while(iter != _errorMessages.end()) {
        std::cerr << iter->getMessage() << std::endl;
        iter++;
      }
      _errorMessages.clear();
      if (_arguments.get_clang_utils()->stopOnAnnotError())
        exit(2);
    }

    while (parseResult == RRContinueLexing)
      parseResult = _arguments.lexer().readToken(buffer, _arguments._position,
          _arguments._tokenLocation, true);
    if (parseResult == RRHasToken) {
      DLexer::AbstractToken::Type type = _arguments.queryToken().getType();
      // This test is to be removed!
      // The preprocessor should be called by the first pass
      //   with the comment handler!
      if (type == DLexer::AbstractToken::TIdentifier) {
        if (_arguments.isPreprocessingDirective(
            ((const DLexer::IdentifierToken&) _arguments.getContentToken())
              .content())) {
          std::string identifier = ((const DLexer::IdentifierToken&)
            _arguments.getContentToken()).content();
          clang::MacroInfo* macro = _arguments.getMacro(identifier);
          if (_arguments._usedMacros.find(identifier)
              != _arguments._usedMacros.end())
            parseResult = RRHasToken;
          else if (_arguments.handleMacroExpandedIdentifier(identifier, macro,
                parseResult)) {
            parseResult = RRNeedChars;
            continue;
          };
        };
      };
      if (parseResult == RRHasToken && type != DLexer::AbstractToken::TComment)
      { do {
          parseResult = _state.parse(_arguments);
          if (_arguments.hasErrors()) {
            std::list<ErrorMessage>& errors = _arguments.errors();
            std::list<ErrorMessage>::const_iterator iterEnd = errors.end();
            for (std::list<ErrorMessage>::const_iterator iter = errors.begin();
                iter != iterEnd; ++iter)
              std::cerr << iter->filepos() << ':' << iter->linepos() << ':'
                << iter->columnpos() << ": " << iter->getMessage() << std::endl;
            errors.clear();
          };
        } while (parseResult == RRContinueParsing);
      }
      else
        parseResult = RRNeedChars;
      if (parseResult == RRNeedChars && _arguments.lexer().doesNeedClear())
        _arguments.lexer().clearToken();
    };
  };
}

} // end of namespace Acsl

